package quimufu.colourful_portals.portal;

import net.minecraft.entity.Entity;
import net.minecraft.entity.projectile.ArrowEntity;
import net.minecraft.entity.projectile.PersistentProjectileEntity;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.TeleportTarget;
import quimufu.colourful_portals.general_util.LinkedList;
import quimufu.colourful_portals.general_util.Node;
import quimufu.colourful_portals.mixin.ServerPlayNetworkHandlerAccessor;
import quimufu.colourful_portals.mixin.ServerPlayerEntityAccessor;
import quimufu.colourful_portals.util.RaycastHelper;

import java.util.*;

import static quimufu.colourful_portals.ColourfulPortalsMod.LOGGER;
import static quimufu.colourful_portals.ColourfulPortalsMod.MOD_ID;
import static quimufu.colourful_portals.portal.PortalHelper.*;

public class DefaultLinkingSystem implements PortalLinkingSystem {

    public static final Identifier DEFAULT_LINKING_SYSTEM = Identifier.of(MOD_ID, "default_linking_system");

    private final Map<BlockPos, Set<Node<PortalRepresentation>>> positionalLookup = new HashMap<>();
    private final MinecraftServer server;

    public DefaultLinkingSystem(MinecraftServer minecraftServer) {
        this.server = minecraftServer;
    }

    @Override
    public Identifier getLinkingSystemId() {
        return DEFAULT_LINKING_SYSTEM;
    }

    @Override
    public void linkPortals(LinkedList<PortalRepresentation> portalRepresentations) {
        LOGGER.debug("start linkPortals");
        for (Node<PortalRepresentation> node = portalRepresentations.getNode(0); node != null; node = node.getNext()) {
            PortalRepresentation portalRepresentation = node.getValue();

            Node<PortalRepresentation> finalNode = node;
            insideOf(portalRepresentation)
                    .forEachRemaining(blockPos -> {
                        Set<Node<PortalRepresentation>> portalsAtPosition = positionalLookup.computeIfAbsent(blockPos, (k) -> new HashSet<>(1));
                        portalsAtPosition.add(finalNode);
                    });
        }
        LOGGER.debug("end linkPortals");
    }

    @Override
    public void unLinkPortal(PortalRepresentation portalRepresentation) {
        //no-op
    }

    @Override
    public boolean movementCallback(Entity entity, BlockPos pos, ServerWorld world) {
        for (Direction.Axis axis : Direction.Axis.values()) {
            Vec3d prevEntityPos;
            if (entity instanceof ServerPlayerEntity serverPlayerEntity) {
                ServerPlayNetworkHandlerAccessor spnha = (ServerPlayNetworkHandlerAccessor) serverPlayerEntity.networkHandler;
                prevEntityPos = new Vec3d(spnha.getLastTickX(), spnha.getLastTickY(), spnha.getLastTickZ());
            } else {
                prevEntityPos = new Vec3d(entity.prevX, entity.prevY, entity.prevZ);
            }
            if (RaycastHelper.passedOnAxis(prevEntityPos, entity.getPos(), pos, axis)) {
                if (onPortalPassed(entity, pos, world, axis)) {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public boolean onPortalPassed(Entity entity, BlockPos pos, ServerWorld world, Direction.Axis a) {
        Set<Node<PortalRepresentation>> portalRepresentationCandidates = positionalLookup.get(pos);
        if (portalRepresentationCandidates == null) {
            return false;
        }
        Optional<Node<PortalRepresentation>> portalOpt = portalRepresentationCandidates.stream()
                .filter(n -> !n.orphaned())
                .filter(n -> n.getValue() != null)
                .filter(n -> getAxisW(n.getValue()).rotateYClockwise().getAxis() == a)
                .filter(n -> Objects.equals(n.getValue().dimensionId(), getDimId(world)))
                .findAny();
        if (portalOpt.isPresent()) {
            Node<PortalRepresentation> fromNode = portalOpt.get();
            Node<PortalRepresentation> toNode = fromNode.getNext() != null ? fromNode.getNext() : fromNode.getPartOf().getNode(0);
            if (toNode == null || toNode.getValue() == null) {
                return false;
            }

            PortalRepresentation fromPortal = fromNode.getValue();
            PortalRepresentation toPortal = toNode.getValue();
            BlockBox fromBox = fromPortal.location();
            Vec3d fromCenter = new Vec3d((fromBox.getMaxX() + 1 + fromBox.getMinX()) / 2D, (fromBox.getMaxY() + 1 + fromBox.getMinY()) / 2D, (fromBox.getMaxZ() + 1 + fromBox.getMinZ()) / 2D);
            BlockBox toBox = toPortal.location();
            Vec3d toCenter = new Vec3d((toBox.getMaxX() + 1 + toBox.getMinX()) / 2D, (toBox.getMaxY() + 1 + toBox.getMinY()) / 2D, (toBox.getMaxZ() + 1 + toBox.getMinZ()) / 2D);
            Vec3d relPos = entity.getPos().subtract(fromCenter);
            Vec3d targetPos;
            float targetYaw;
            Vec3d targetVelocity;
            ServerWorld toWorld = getPortalWorld(server, toPortal);
            if (getAxisW(fromPortal) == getAxisW(toPortal)) {
                targetPos = toCenter.add(relPos);
                targetYaw = entity.getYaw();
                targetVelocity = entity.getVelocity();
            } else {
                targetPos = toCenter.add(new Vec3d(-relPos.z, relPos.y, relPos.x));
                if (entity instanceof PersistentProjectileEntity) {
//                    //un-fuck up arrow Yaw
//                    //rotate by 90°
//                    targetYaw = (entity.getYaw() + 90.F) % 360.F;
//                    //mirror
//                    targetYaw = (-targetYaw) % 360.F;
//                    //rotate by 90°
//                    targetYaw = (targetYaw - 90.F) % 360.F;
//
//                    //now it's normal!
//                    //Let's rotate it!
//                    targetYaw = (targetYaw + 90.F) % 360.F;
//
//                    //fuck it up again
//                    //rotate by 90°
//                    targetYaw = (targetYaw + 90.F) % 360.F;
//                    //mirror
//                    targetYaw = (-targetYaw) % 360.F;
//                    //rotate by 90°
//                    targetYaw = (targetYaw - 90.F) % 360.F;
                    //short:
                    targetYaw = (entity.getYaw() - 90.F) % 360.F;
                } else {
                    targetYaw = (entity.getYaw() + 90.F) % 360.F;
                }
                Vec3d currVelocity = entity.getVelocity();
                targetVelocity = new Vec3d(-currVelocity.z, currVelocity.y, currVelocity.x);
            }
            //todo: maybe fancy continoous movement math!
            TeleportTarget teleportTarget = new TeleportTarget(toWorld, targetPos, targetVelocity, targetYaw, entity.getPitch(), TeleportTarget.ADD_PORTAL_CHUNK_TICKET);
            if (entity instanceof ServerPlayerEntity) {
                ((ServerPlayerEntityAccessor) entity).setInTeleportationState(true);
            }
            Entity target;
            if ((target = entity.teleportTo(teleportTarget)) != null) {
                target.velocityDirty = true;
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean needsReInit() {
        return true;
    }

    @Override
    public boolean needsMovementCallback() {
        return true;
    }

    private static ServerWorld getPortalWorld(MinecraftServer server, PortalRepresentation fromPortalRepresentation) {
        ServerWorld serverWorld = PortalHelper.getPortalWorld(server, fromPortalRepresentation);
        if (serverWorld == null) {
            LOGGER.warn("couldn't get portal dimensionId for portal {}. Don't sue me!", fromPortalRepresentation);
            throw new RuntimeException();
        }
        return serverWorld;
    }
}
