package quimufu.colourful_portals.util;

import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.entity.Entity;
import net.minecraft.entity.LivingEntity;
import net.minecraft.entity.mob.MobEntity;
import net.minecraft.registry.Registries;
import net.minecraft.registry.RegistryKey;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.ServerTask;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.server.world.ChunkTicketType;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.sound.SoundCategory;
import net.minecraft.sound.SoundEvents;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.*;
import net.minecraft.world.Heightmap;
import net.minecraft.world.TeleportTarget;
import net.minecraft.world.World;
import net.minecraft.world.border.WorldBorder;
import net.minecraft.world.chunk.Chunk;
import net.minecraft.world.dimension.DimensionType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.joml.Random;
import org.joml.Vector3f;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import quimufu.colourful_portals.ColourfulPortalsMod;
import quimufu.colourful_portals.config.ColourfulPortalConfig;
import quimufu.colourful_portals.mixin.ServerWorldAccessor;
import quimufu.colourful_portals.portal.PortalHelper;

import java.util.Comparator;
import java.util.Iterator;
import java.util.UUID;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

public class TeleportHelper extends ServerTask {

    private static final Logger log = LoggerFactory.getLogger(TeleportHelper.class);
    public static final int TICK_DELAY = 2;
    private final Random random;
    private final MinecraftServer minecraftServer;
    private final ConcurrentSkipListSet<TeleportRequest> toTeleport = new ConcurrentSkipListSet<>();
    private static TeleportHelper INSTANCE;
    private final AtomicBoolean tickScheduled = new AtomicBoolean(false);
    private int lastExecTick;
    private final ScheduledExecutorService sleeperExecutor = Executors.newSingleThreadScheduledExecutor();

    private static TeleportHelper getInstance(MinecraftServer server) {
        if (INSTANCE == null) {
            INSTANCE = new TeleportHelper(server);
        }
        return INSTANCE;
    }

    private TeleportHelper(MinecraftServer server) {
        super(server.getTicks(), null);
        this.minecraftServer = server;
        this.random = new Random();
    }


    private boolean targetValid(Entity entity, ServerWorld world, Vector3f targetPos, boolean force) {
        Box boundingBox = entity.getDimensions(entity.getPose())
                .getBoxAt(new Vec3d(targetPos))
                .stretch(0, -1, 0);
        boolean foundAir = false;
        boolean foundNonAir = false;
        for (BlockPos pos : PortalHelper.blockPosInBox(boundingBox)) {
            BlockState blockState = world.getBlockState(pos);
            if (!blockState.isIn(ColourfulPortalsMod.COLOURFUL_PEARL_REPLACEABLE_BLOCK_TAG)) {
                ColourfulPortalsMod.LOGGER.info("invalid location at {}, block {}", pos, Registries.BLOCK.getId(blockState.getBlock()));
                return false;
            }
            if (blockState.isAir()) {
                foundAir = true;
            } else {
                foundNonAir = true;
            }
        }
        return (foundNonAir) && (foundAir || force);
    }

    private @Nullable TeleportTarget getTarget(Entity entity, ServerWorld serverWorld, Vector3f from, ServerWorld targetWorld, BlockPos targetBlockPos) {
        minecraftServer.getProfiler().push("getChunk");
        Chunk chunk = targetWorld.getChunk(targetBlockPos);
        minecraftServer.getProfiler().pop();

        if (isEmptyPosition(targetWorld, targetBlockPos, chunk)) {
            log.info("empty at {}, retrying", targetBlockPos);
            TeleportHelper.markForTeleport(entity, from, serverWorld);
            return null;
        }
        Vector3f targetPos = Vec3d.of(targetBlockPos).toVector3f();
        int width = MathHelper.ceil(entity.getBoundingBox().getLengthX());
        int depth = MathHelper.ceil(entity.getBoundingBox().getLengthZ());
        targetPos.add((width % 2) * 0.5F, 0, (depth % 2) * 0.5F);
        minecraftServer.getProfiler().push("valid check");
        int tries = 0;
        while (!targetValid(entity, targetWorld, targetPos, tries > targetWorld.getLogicalHeight() * 2)) {
            if (tries > targetWorld.getLogicalHeight() * 4) {
                break;
            }
            targetPos.y = getTargetY(targetWorld);
            tries++;
        }
        minecraftServer.getProfiler().pop();
        return new TeleportTarget(targetWorld, new Vec3d(targetPos), entity.getVelocity(), entity.getYaw(), entity.getPitch(), TeleportTarget.NO_OP);
    }

    private boolean isEmptyPosition(ServerWorld world, BlockPos target, Chunk chunk) {
        minecraftServer.getProfiler().push("sampleHeightmap");
        int height = chunk.sampleHeightmap(Heightmap.Type.WORLD_SURFACE, target.getX() & 0xF, target.getZ() & 0xF) + 1;
        minecraftServer.getProfiler().pop();
        return height == world.getBottomY();
    }

    private @Nullable ServerWorld getTargetWorld(ServerWorld serverWorld) {
        if (random.nextFloat() < ColourfulPortalConfig.pearlSameDimensionLikelihood) {
            return serverWorld;
        }

        MinecraftServer server = serverWorld.getServer();
        ColourfulPortalConfig.addMissingDimensionsToConfig(server);
        Identifier targetWorldId = ColourfulPortalsMod.DIMENSION_WEIGHTS_COLOURFUL_PEARL.getWeighted(random);

        return server.getWorld(RegistryKey.of(RegistryKeys.WORLD, targetWorldId));
    }

    private BlockPos getTargetPos(ServerWorld fromWorld, ServerWorld toWorld, Vector3f pos) {

        double distance = ColourfulPortalConfig.minPearlDistance +
                          random.nextFloat() *
                          (ColourfulPortalConfig.maxPearlDistance - ColourfulPortalConfig.minPearlDistance);

        double angle = Math.PI * 2 * random.nextFloat();
        int targetY = getTargetY(toWorld);

        Vec3d target = new Vec3d(pos.x + Math.cos(angle) * distance, targetY, pos.z + Math.sin(angle) * distance);

        WorldBorder worldBorder = toWorld.getWorldBorder();
        double d = DimensionType.getCoordinateScaleFactor(fromWorld.getDimension(), toWorld.getDimension());
        return worldBorder.clamp(target.getX() * d, target.getY(), target.getZ() * d);
    }

    private int getTargetY(ServerWorld toWorld) {
        return toWorld.getBottomY() + 4 + (random.nextInt(toWorld.getLogicalHeight() - 8));
    }


    public void tryTeleport(Entity entity, Vector3f from, ServerWorld serverWorld, BlockPos to, ServerWorld toWorld) {
        if (entity == null) {
            return;
        }
        if (entity.hasVehicle()) {
            entity.detach();
        }
        minecraftServer.getProfiler().push("getTarget");
        TeleportTarget target = getTarget(entity, serverWorld, from, toWorld, to);
        minecraftServer.getProfiler().pop();
        if (target == null) {
            return;
        }
        prepareTarget(target, entity);
        ((ServerWorldAccessor) serverWorld).getEntityList()
                .add(entity);

        minecraftServer.getProfiler().push("teleportTo");
        if (entity instanceof ServerPlayerEntity serverPlayerEntity) {
            if (serverPlayerEntity.networkHandler.isConnectionOpen()) {

                Entity entityAfterTeleport = entity.teleportTo(target);
                if (entityAfterTeleport == null) {
                    minecraftServer.getProfiler().pop();
                    return;
                }
                entityAfterTeleport.onLanding();
                serverPlayerEntity.clearCurrentExplosion();
                entityAfterTeleport.damage(entityAfterTeleport.getDamageSources().fall(), 5.0f);
                serverPlayerEntity.playSoundToPlayer(SoundEvents.ENTITY_PLAYER_TELEPORT, SoundCategory.PLAYERS, 1F, 0.75F);
            }
        } else {
            Entity entityAfterTeleport = entity.teleportTo(target);
            if (entityAfterTeleport == null) {
                minecraftServer.getProfiler().pop();
                return;
            }
            entityAfterTeleport.onLanding();
            serverWorld.playSound(entityAfterTeleport, entityAfterTeleport.getBlockPos(),
                    SoundEvents.ENTITY_PLAYER_TELEPORT, SoundCategory.NEUTRAL,
                    1F, 0.75F);
        }
        minecraftServer.getProfiler().pop();
    }


    private void prepareTarget(TeleportTarget target, Entity entity) {
        Box entityBody = entity.getDimensions(entity.getPose()).getBoxAt(target.pos());
        double minY = entityBody.getMinPos().getY();
        Box floor = entityBody.withMinY(minY - 1).withMaxY(minY - 0.5);
        for (BlockPos pos : PortalHelper.blockPosInBox(entityBody)) {
            target.world().setBlockState(pos, ColourfulPortalsMod.COLOURFUL_AIR.getRandomState(random));
        }
        for (BlockPos pos : PortalHelper.blockPosInBox(floor)) {
            target.world().setBlockState(pos, Blocks.STONE.getDefaultState());
        }

    }

    private static boolean canTeleportEntityTo(Entity entity, World world) {
        if (entity.getWorld().getRegistryKey() == world.getRegistryKey()) {
            if (entity instanceof LivingEntity livingEntity) {
                return livingEntity.isAlive() && !livingEntity.isSleeping();
            }
            return entity.isAlive();
        }
        return entity.canUsePortals(true);
    }


    private void addTeleportRequest(Entity entity, Vector3f from,
                                    ServerWorld fromWorld,
                                    boolean originalInvulnerability) {
        ServerWorld targetWorld = getTargetWorld(fromWorld);
        if (targetWorld == null) {
            log.info("couldn't find target world");
            return;
        }
        BlockPos targetBlockPos = getTargetPos(fromWorld, targetWorld, from);
        if (entity instanceof MobEntity mobEntity
            && !mobEntity.isPersistent()
            && !mobEntity.cannotDespawn()
            && mobEntity.canImmediatelyDespawn(targetBlockPos.getSquaredDistance(new Vec3d(from)))
            && !targetWorld.isChunkLoaded(targetBlockPos)) {
            log.info("despawning {} instead of teleporting it OOB", mobEntity.getName().getString());

            ((ServerWorldAccessor) fromWorld).getEntityList().add(entity);
            mobEntity.discard();
            return;
        }
        if (!canTeleportEntityTo(entity, targetWorld)) {
            return;
        }
        targetWorld.getChunkManager()
                .addTicket(ChunkTicketType.PORTAL, new ChunkPos(targetBlockPos), 2, targetBlockPos);
        toTeleport.add(new TeleportRequest(entity.getUuid(),
                from,
                fromWorld.getRegistryKey().getValue(),
                originalInvulnerability,
                targetBlockPos,
                targetWorld.getRegistryKey().getValue()
        ));

        if (!tickScheduled.getAndSet(true)) {
            long nanosPerTick = minecraftServer.getTickManager().getNanosPerTick();
            sleeperExecutor.schedule(() -> minecraftServer.send(this), nanosPerTick * 16, TimeUnit.NANOSECONDS);
        }
    }

    @Override
    public void run() {
        log.info("{}", minecraftServer.getTicks());
        int nextExecIn = (lastExecTick + TICK_DELAY) - minecraftServer.getTicks();
        long nanosPerTick = minecraftServer.getTickManager().getNanosPerTick();
        if (nextExecIn > 0) {
            minecraftServer.send(this);
            return;
        }
        this.lastExecTick = minecraftServer.getTicks();
        minecraftServer.getProfiler().push("teleportLoaded");
        teleportLoaded();
        minecraftServer.getProfiler().pop();
        log.info("there are currently {} entities awaiting teleportation", toTeleport.size());
        if (!toTeleport.isEmpty()) {
            minecraftServer.send(this);
        } else {
            tickScheduled.set(false);
        }
    }


    private void teleportLoaded() {
        Iterator<TeleportRequest> iterator = toTeleport.iterator();
        while (iterator.hasNext()) {
            TeleportRequest teleportRequest = iterator.next();
            if (teleportRequest == null) {
                //empty for some reason!
                iterator.remove();
                log.error("null teleportRequest");
                return;
            }
            ServerWorld fromWorld = minecraftServer.getWorld(RegistryKey.of(RegistryKeys.WORLD, teleportRequest.fromWorldId));
            if (fromWorld == null) {
                iterator.remove();
                log.error("lost from world");
                return;
            }
            ServerWorld toWorld = minecraftServer.getWorld(RegistryKey.of(RegistryKeys.WORLD, teleportRequest.toWorldId));
            if (toWorld == null) {
                iterator.remove();
                log.error("lost to world");
                return;
            }
            Entity entity = fromWorld.getEntity(teleportRequest.entity);
            if (entity == null) {
                iterator.remove();
                log.error("lost entity");
                return;
            }
            int x = teleportRequest.to.getX();
            int z = teleportRequest.to.getZ();
            if (toWorld.getChunkManager().isChunkLoaded(ChunkSectionPos.getSectionCoord(x), ChunkSectionPos.getSectionCoord(z))) {
                log.info("chunkLoaded {}", teleportRequest.to);
                minecraftServer.getProfiler().push("tryTeleport");
                tryTeleport(entity, teleportRequest.from, fromWorld, teleportRequest.to, toWorld);
                minecraftServer.getProfiler().pop();
                iterator.remove();
                return;
            } else {
                toWorld.getChunkManager()
                        .addTicket(ChunkTicketType.PORTAL, new ChunkPos(teleportRequest.to), 2, teleportRequest.to);
            }
        }

    }

    public static void markForTeleport(Entity entity, Vector3f from, ServerWorld serverWorld) {
        ((ServerWorldAccessor) serverWorld).getEntityList()
                .remove(entity);
        boolean originalInvulnerability = entity.isInvulnerable();
        entity.setInvulnerable(true);
        getInstance(serverWorld.getServer())
                .addTeleportRequest(entity, from, serverWorld, originalInvulnerability);

    }

    private record TeleportRequest(UUID entity, Vector3f from,
                                   Identifier fromWorldId,
                                   boolean originalInvulnerability,
                                   BlockPos to,
                                   Identifier toWorldId) implements Comparable<TeleportRequest> {
        @Override
        public int compareTo(@NotNull TeleportHelper.TeleportRequest other) {
            Comparator<TeleportRequest> comparator = Comparator
                    .comparing((TeleportRequest teleportRequest) -> teleportRequest.entity);
            return comparator.compare(this, other);
        }
    }
}
